/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_interactive_manipulation/advanced_options_dialog.h"
#include "pr2_interactive_manipulation/interactive_manipulation_frontend.h"

namespace pr2_interactive_manipulation {

AdvancedOptionsDialog::AdvancedOptionsDialog(InteractiveManipulationFrontend *frontend) :
  AdvancedOptionsDialogBase(frontend),
  frontend_(frontend)
{
}

AdvancedOptionsDialog::~AdvancedOptionsDialog()
{
}

void AdvancedOptionsDialog::setDefaultsClicked( wxCommandEvent& )
{
  setOptions( getDefaults() );
}

AdvancedGraspOptions AdvancedOptionsDialog::getDefaults(int interface_number, int task_number)
{
  AdvancedGraspOptions go;
  go.reactive_grasping_ = false;
  go.reactive_force_ = false;
  go.reactive_place_ = false;
  go.lift_steps_  = 10;
  go.retreat_steps_ = 10;
  go.lift_direction_choice_ = 0;
  go.desired_approach_ = 10;
  go.min_approach_ = 5;
  go.max_contact_force_ = 50;
  switch (task_number)
  {
  case 3:
    go.lift_direction_choice_ = 1;
    break;
  }

  return go;
}

AdvancedGraspOptions AdvancedOptionsDialog::getOptions()
{
  AdvancedGraspOptions go;
  go.reactive_grasping_ = reactive_grasping_box_->IsChecked();
  go.reactive_force_ = reactive_transport_box_->IsChecked();
  go.reactive_place_ = reactive_place_box_->IsChecked();
  go.lift_steps_  = lift_distance_spin_->GetValue();
  go.retreat_steps_ = retreat_distance_spin_->GetValue();
  go.lift_direction_choice_ = lift_direction_choice_->GetSelection();
  go.desired_approach_ = desired_grasp_approach_spin_->GetValue();
  go.min_approach_ = min_grasp_approach_spin_->GetValue();
  go.max_contact_force_ = max_contact_force_spin_->GetValue();
  return go;
}

void AdvancedOptionsDialog::setOptions(AdvancedGraspOptions go)
{
  reactive_grasping_box_->SetValue(go.reactive_grasping_);
  reactive_transport_box_->SetValue(go.reactive_force_);
  reactive_place_box_->SetValue(go.reactive_place_);
  lift_distance_spin_->SetValue(go.lift_steps_);
  retreat_distance_spin_->SetValue(go.retreat_steps_);
  lift_direction_choice_->SetSelection(go.lift_direction_choice_);
  desired_grasp_approach_spin_->SetValue(go.desired_approach_);
  min_grasp_approach_spin_->SetValue(go.min_approach_);
  max_contact_force_spin_->SetValue(go.max_contact_force_);
}

void AdvancedOptionsDialog::acceptClicked( wxCommandEvent& event )
{
  frontend_->setAdvancedOptions(getOptionsMsg());
  Close();
}


pr2_object_manipulation_msgs::IMGUIAdvancedOptions AdvancedOptionsDialog::getDefaultsMsg(int interface_number, int task_number)
{
  pr2_object_manipulation_msgs::IMGUIAdvancedOptions go;
  go.reactive_grasping = false;
  go.reactive_force = false;
  go.reactive_place = false;
  go.lift_steps  = 10;
  go.retreat_steps = 10;
  go.lift_direction_choice = 0;
  go.desired_approach = 10;
  go.min_approach = 5;
  go.max_contact_force = 50;
  switch (task_number)
  {
  case 3:
    go.lift_direction_choice = 1;
    break;
  }
  return go;
}

pr2_object_manipulation_msgs::IMGUIAdvancedOptions AdvancedOptionsDialog::getOptionsMsg()
{
  pr2_object_manipulation_msgs::IMGUIAdvancedOptions go;
  go.reactive_grasping = reactive_grasping_box_->IsChecked();
  go.reactive_force = reactive_transport_box_->IsChecked();
  go.reactive_place = reactive_place_box_->IsChecked();
  go.lift_steps  = lift_distance_spin_->GetValue();
  go.retreat_steps = retreat_distance_spin_->GetValue();
  go.lift_direction_choice = lift_direction_choice_->GetSelection();
  go.desired_approach = desired_grasp_approach_spin_->GetValue();
  go.min_approach = min_grasp_approach_spin_->GetValue();
  go.max_contact_force = max_contact_force_spin_->GetValue();
  return go;
}

void AdvancedOptionsDialog::setOptionsMsg(pr2_object_manipulation_msgs::IMGUIAdvancedOptions go)
{
  reactive_grasping_box_->SetValue(go.reactive_grasping);
  reactive_transport_box_->SetValue(go.reactive_force);
  reactive_place_box_->SetValue(go.reactive_place);
  lift_distance_spin_->SetValue(go.lift_steps);
  retreat_distance_spin_->SetValue(go.retreat_steps);
  lift_direction_choice_->SetSelection(go.lift_direction_choice);
  desired_grasp_approach_spin_->SetValue(go.desired_approach);
  min_grasp_approach_spin_->SetValue(go.min_approach);
  max_contact_force_spin_->SetValue(go.max_contact_force);
}

void AdvancedOptionsDialog::reactiveGraspingClicked( wxCommandEvent& )
{
  //reactive gripper force is only allowed together with reactive grasping
  if (!reactive_grasping_box_->IsChecked()) reactive_transport_box_->SetValue(false);
}

void AdvancedOptionsDialog::reactiveForceClicked( wxCommandEvent& )
{
  //reactive gripper force is only allowed together with reactive grasping
  if (!reactive_grasping_box_->IsChecked()) reactive_grasping_box_->SetValue(true);  
}

}
