/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_interactive_manipulation/camera_focus_display.h"

#include <rviz/window_manager_interface.h>
#include <rviz/visualization_manager.h>
#include <rviz/view_controller.h>
#include <rviz/frame_manager.h>

#include <OGRE/OgreMatrix3.h>

#include <object_manipulator/tools/camera_configurations.h>

namespace pr2_interactive_manipulation {

CameraFocusDisplay::CameraFocusDisplay( const std::string& name, 
                                        rviz::VisualizationManager* manager ) : 
  Display( name, manager ),
  visualization_manager_(manager),
  name_(name),
  frame_(NULL)
{
  window_manager_ = visualization_manager_->getWindowManager();
  ROS_ASSERT(window_manager_);
  wxWindow* parent = window_manager_->getParentWindow();
  ROS_ASSERT(parent);
}
  
CameraFocusDisplay::~CameraFocusDisplay()
{
  if ( frame_ )
  {
    window_manager_->removePane( frame_ );
    delete frame_;
  }
}

void CameraFocusDisplay::onEnable()
{  
  if ( !frame_ )
  {
    frame_ = new CameraFocusFrame(window_manager_->getParentWindow(), vis_manager_);
    window_manager_->addPane( "Camera Focus", frame_ );
  }
  window_manager_->showPane(frame_); 
}

void CameraFocusDisplay::onDisable()
{
  if ( frame_) window_manager_->closePane(frame_);
}

void CameraFocusDisplay::createProperties()
{
}

void CameraFocusDisplay::update(float, float)
{
}


CameraFocusFrame::CameraFocusFrame(wxWindow *parent,
                                   rviz::VisualizationManager* manager) :
  CameraFocusFrameBase(parent),
  root_nh_(""),
  vis_manager_(manager)
{
  sub_ = root_nh_.subscribe("/camera_focus", 10, &CameraFocusFrame::callback, this);    
}

CameraFocusFrame::~CameraFocusFrame()
{
}

void CameraFocusFrame::callback(const pr2_object_manipulation_msgs::CameraFocusConstPtr &msg)
{
  ROS_DEBUG("Camera focus message received");
  if (!accept_box_->IsChecked()) {ROS_INFO("Camera focus: not accepting external commands"); return;}

  //transform point into target frame
  geometry_msgs::PointStamped point = msg->focal_point;
  point.header.stamp = ros::Time(0);
  try {
    vis_manager_->getFrameManager()->getTFClient()->transformPoint(vis_manager_->getTargetFrame(), 
                                                                   point, point );
  }
  catch(...)
  {
    ROS_ERROR("Camera focus: TF had a problem transforming between [%s] and [%s].",
              point.header.frame_id.c_str(), vis_manager_->getTargetFrame().c_str());
    return;
  }
  std::string params_string = vis_manager_->getCurrentViewController()->toString();
  ROS_DEBUG_STREAM("View params: " << params_string);

  //get the current params
  std::istringstream params_str(params_string); 
  std::vector<double> params;
  for (size_t i=0; i<6; i++)
  {
    double p;
    params_str >> p;
    if (params_str.fail())
    {
      ROS_ERROR("Camera focus: could not interpret params string");
      return;
    }
    params.push_back(p);
  }

  //set the desired point
  //arbitrary zoom at 1 meter
  params[2] = 1.5;
  //there is some funky axis re-mapping going on here
  params[3] = -point.point.y;
  params[4] =  point.point.z;
  params[5] = -point.point.x;

  //set the camera params
  //setCamera(params);
  /*
  std::ostringstream os;
  for(int i=0; i<6; i++) os << params[i] << ' ';
  vis_manager_->getCurrentViewController()->fromString(os.str());
  vis_manager_->queueRender();
  ROS_DEBUG("Camera focus set");
  */

  float yaw_correction = 0;
  Ogre::Vector3 pos;
  Ogre::Quaternion orient;

  //correct for yaw rotation
  if (rviz::FrameManager::instance()->getTransform("base_link", 
						    ros::Time(),  pos, orient) )
  {
    yaw_correction = orient.getRoll().valueRadians();
  }
  //params[1] -= yaw_correction;

  //also need to rotate the focal point
  Ogre::Vector3 focal_point(params[3], params[4], params[5]);
  Ogre::Matrix3 rot;
  rot.FromAxisAngle(Ogre::Vector3(0,1,0), Ogre::Radian(yaw_correction));
  focal_point = rot * focal_point;
  for(int i=0; i<3; i++) params[3+i] = focal_point[i];

  //set the camera params
  std::ostringstream os;
  for(int i=0; i<6; i++) os << params[i] << ' ';
  vis_manager_->setTargetFrame("base_link");
  vis_manager_->setCurrentViewControllerType( "Orbit" );
  vis_manager_->getCurrentViewController()->fromString(os.str());
  vis_manager_->queueRender();
}


void CameraFocusFrame::setCamera(std::vector<double> params)
{
  float yaw_correction = 0;
  Ogre::Vector3 pos;
  Ogre::Quaternion orient;

  //correct for yaw rotation
  if (rviz::FrameManager::instance()->getTransform("base_link", 
						    ros::Time(),  pos, orient) )
  {
    yaw_correction = orient.getRoll().valueRadians();
  }
  params[1] -= yaw_correction;

  //also need to rotate the focal point
  Ogre::Vector3 focal_point(params[3], params[4], params[5]);
  Ogre::Matrix3 rot;
  rot.FromAxisAngle(Ogre::Vector3(0,1,0), Ogre::Radian(yaw_correction));
  focal_point = rot * focal_point;
  for(int i=0; i<3; i++) params[3+i] = focal_point[i];

  //set the camera params
  std::ostringstream os;
  for(int i=0; i<6; i++) os << params[i] << ' ';
  vis_manager_->setTargetFrame("base_link");
  vis_manager_->setCurrentViewControllerType( "Orbit" );
  vis_manager_->getCurrentViewController()->fromString(os.str());
  vis_manager_->queueRender();
}

void CameraFocusFrame::leftButtonClicked( wxCommandEvent&  )
{
  try
  {
    setCamera( object_manipulator::cameraConfigurations().get_camera_pose("left") );
  }
  catch(object_manipulator::BadParamException &ex)
  {
    ROS_ERROR("Pre-defined camera position not found");
  }
}

void CameraFocusFrame::topButtonClicked( wxCommandEvent&  )
{
  try
  {
    setCamera( object_manipulator::cameraConfigurations().get_camera_pose("top") );
  }
  catch(object_manipulator::BadParamException &ex)
  {
    ROS_ERROR("Pre-defined camera position not found");
  }
}

void CameraFocusFrame::frontButtonClicked( wxCommandEvent& )
{
  try
  {
    setCamera( object_manipulator::cameraConfigurations().get_camera_pose("front") );
  }
  catch(object_manipulator::BadParamException &ex)
  {
    ROS_ERROR("Pre-defined camera position not found");
  }
}

void CameraFocusFrame::rightButtonClicked( wxCommandEvent&  )
{
  try
  {
    setCamera( object_manipulator::cameraConfigurations().get_camera_pose("right") );
  }
  catch(object_manipulator::BadParamException &ex)
  {
    ROS_ERROR("Pre-defined camera position not found");
  }
}

void CameraFocusFrame::overheadButtonClicked( wxCommandEvent&  )
{
  try
  {
    setCamera( object_manipulator::cameraConfigurations().get_camera_pose("overhead") );
  }
  catch(object_manipulator::BadParamException &ex)
  {
    ROS_ERROR("Pre-defined camera position not found");
  }
}

void CameraFocusFrame::acceptBoxChecked( wxCommandEvent&  )
{
}

}
