/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "pr2_interactive_manipulation/interactive_manipulation_frontend.h"

#include "object_manipulator/tools/camera_configurations.h"

#include <rviz/visualization_manager.h>
#include <rviz/view_controller.h>
#include <rviz/frame_manager.h>
#include <OGRE/OgreMatrix3.h>

namespace pr2_interactive_manipulation {

typedef actionlib::SimpleActionClient<pr2_object_manipulation_msgs::IMGUIAction> Client;

InteractiveManipulationFrontend::InteractiveManipulationFrontend(wxWindow *parent, 
                                                                 rviz::VisualizationManager* manager) : 
  InteractiveManipulationFrameBase(parent),
  vis_manager_(manager),
  root_nh_(""),
  status_label_text_("idle"),
  adv_options_(AdvancedOptionsDialog::getDefaultsMsg())
{
  root_nh_.param<int>("interactive_grasping/interface_number", interface_number_, 0);
  if (!interface_number_) ROS_WARN("No interface number specified for grasping study; using default configuration");
  else ROS_INFO("Using interface number %d for grasping study", interface_number_);
  root_nh_.param<int>("interactive_grasping/task_number", task_number_, 0);
  if (!task_number_) ROS_WARN("No task number specified for grasping study; using default configuration");
  else ROS_INFO("Using task number %d for grasping study", task_number_);

  action_name_ = "imgui_action";  
  action_client_ = new actionlib::SimpleActionClient<pr2_object_manipulation_msgs::IMGUIAction>(action_name_, true);

  status_name_ = "interactive_manipulation_status";
  status_sub_ = root_nh_.subscribe(status_name_, 1, &InteractiveManipulationFrontend::statusCallback, this);

  rcommander_action_info_name_ = "list_rcommander_actions";
  rcommander_group_name_ = "house_hold";
  rcommander_action_info_client_ = root_nh_.serviceClient<pr2_object_manipulation_msgs::ActionInfo>(rcommander_action_info_name_);
  rcommanderRefresh();

  //scripted_action_name_ = "run_rcommander_action";
  //scripted_action_client_ = new actionlib::SimpleActionClient<RunScriptAction>(scripted_action_name, true);

  collision_panel_->Enable(false);
  helper_panel_->Enable(false);
  actions_panel_->Enable(false);
  place_button_->Enable(false);
  stop_nav_button_->Enable(true);

  advanced_options_button_->Enable(false);
  adv_options_ = AdvancedOptionsDialog::getDefaultsMsg(interface_number_, task_number_);

  switch (interface_number_)
  {
  case 1:
    collision_box_->SetValue(false);
    grasp_button_->Enable(false);
    planned_move_button_->Enable(false);
    break;
  case 2:
    collision_box_->SetValue(false);
    grasp_button_->Enable(false);
    planned_move_button_->Enable(true);
    break;
  case 3:
    grasp_button_->Enable(true);
    planned_move_button_->Enable(false);
    break;
  case 4:
    grasp_button_->Enable(true);
    planned_move_button_->Enable(false);
    break;
  case 0:
    collision_panel_->Enable(true);
    helper_panel_->Enable(true);
    actions_panel_->Enable(true);
    place_button_->Enable(true);
    advanced_options_button_->Enable(true);
  } 
  grasp_place_panel_->SetFocus();
}

InteractiveManipulationFrontend::~InteractiveManipulationFrontend()
{
  delete action_client_;
}

wxString getWxString(std::string str)
{
  wxString mystring(str.c_str(), wxConvUTF8);
  return mystring;
}

void InteractiveManipulationFrontend::update()
{
  status_label_mutex_.lock();
  wxString mystring(status_label_text_.c_str(), wxConvUTF8);  
  status_label_mutex_.unlock();
  status_label_->SetLabel(mystring);

  reset_button_->Enable(collision_box_->IsChecked());
  reset_choice_->Enable(collision_box_->IsChecked());
}

void InteractiveManipulationFrontend::statusCallback( const std_msgs::StringConstPtr &status)
{
  ROS_DEBUG_STREAM("IM Frontend received stauts: " << status->data);
  boost::mutex::scoped_lock lock(status_label_mutex_);
  status_label_text_ = status->data;
}

void InteractiveManipulationFrontend::feedbackCallback(const pr2_object_manipulation_msgs::IMGUIFeedbackConstPtr &feedback)
{
  //we are now doing this with topic subscription instead
  /*
  ROS_DEBUG_STREAM("IM Frontend received feedback" << feedback->status);
  boost::mutex::scoped_lock lock(status_label_mutex_);
  status_label_text_ = feedback->status;
  */  
}

void InteractiveManipulationFrontend::cancelButtonClicked( wxCommandEvent& /*event*/ )
{
  action_client_->cancelAllGoals();
}

void InteractiveManipulationFrontend::stopNavButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.STOP_NAV;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::graspButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.PICKUP;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::placeButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.PLACE;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::plannedMoveButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.PLANNED_MOVE;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::resetButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.RESET;  
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::armGoButtonClicked( wxCommandEvent& /*event*/ )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.MOVE_ARM;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::modelObjectClicked( wxCommandEvent& /*event*/)
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.MODEL_OBJECT;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

//void InteractiveManipulationFrontend::openDrawerButtonClicked( wxCommandEvent& event)
//{
//  IMGUIGoal goal;
//  goal.options = getDialogOptions();
//  goal.command.command = goal.command.OPEN_DRAWER;
//  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
//                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
//}

void InteractiveManipulationFrontend::rcommandRunButtonClicked( wxCommandEvent& event)
{
    //Call RunScriptAction
    int selection;
    pr2_object_manipulation_msgs::IMGUIGoal goal;
    selection = rcommander_choice->GetSelection();
    if (selection == -1)
        return;

    goal.options = getDialogOptions();
    goal.command.command = goal.command.SCRIPTED_ACTION;
    goal.command.script_name = std::string(rcommander_choice->GetString(selection).mb_str());
    goal.command.script_group_name = rcommander_group_name_;
    action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                             boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::rcommandRefreshButtonClicked( wxCommandEvent& event)
{
    rcommanderRefresh();
}

void InteractiveManipulationFrontend::rcommanderRefresh()
{
    //Call service for list of actions
    pr2_object_manipulation_msgs::ActionInfo srv;
    srv.request.group_name = rcommander_group_name_;
    if (rcommander_action_info_client_.call(srv))
    {
        rcommander_choice->Clear();
        for (unsigned int i = 0; i < srv.response.services.size(); i++)
        {
            rcommander_choice->Append(getWxString(srv.response.services[i]));
            //rcommander_choice->Append(wxT(srv.response.services[i].c_str()));
        }
	    rcommander_choice->SetSelection( 0 );
    }
}


void InteractiveManipulationFrontend::gripperSliderScrollChanged( wxScrollEvent& )
{
  pr2_object_manipulation_msgs::IMGUIGoal goal;
  goal.options = getDialogOptions();
  goal.command.command = goal.command.MOVE_GRIPPER;
  action_client_->sendGoal(goal,Client::SimpleDoneCallback(), Client::SimpleActiveCallback(),
                           boost::bind(&InteractiveManipulationFrontend::feedbackCallback, this, _1));
}

void InteractiveManipulationFrontend::advancedOptionsClicked( wxCommandEvent& /*event*/ )
{
  AdvancedOptionsDialog *dlg = new AdvancedOptionsDialog(this);
  dlg->setOptionsMsg(adv_options_);
  dlg->Show();
}

pr2_object_manipulation_msgs::IMGUIOptions InteractiveManipulationFrontend::getDialogOptions()
{
  pr2_object_manipulation_msgs::IMGUIOptions options;
  options.collision_checked = collision_box_->IsChecked();
  options.grasp_selection = 0;
  options.arm_selection = arm_choice_->GetSelection();
  options.reset_choice = reset_choice_->GetSelection();
  options.arm_action_choice = arm_action_choice_->GetSelection();
  options.arm_planner_choice = arm_planner_choice_->GetSelection();
  options.adv_options = adv_options_;
  options.gripper_slider_position = gripper_slider_->GetValue();
  return options;
}

//rviz has an annoying relative transform that makes changing the target frame 
//destroy the view, which will be fixed eventually
//if you keep the target frame in base_link and don't change it, this should
//continue to give you views relative to the robot
void InteractiveManipulationFrontend::setCamera(std::vector<double> params)
{
  float yaw_correction = 0;
  Ogre::Vector3 pos;
  Ogre::Quaternion orient;

  //correct for yaw rotation
  if (rviz::FrameManager::instance()->getTransform("base_link", 
						    ros::Time(),  pos, orient) )
  {
    yaw_correction = orient.getRoll().valueRadians();
  }
  params[1] -= yaw_correction;

  //also need to rotate the focal point
  Ogre::Vector3 focal_point(params[3], params[4], params[5]);
  Ogre::Matrix3 rot;
  rot.FromAxisAngle(Ogre::Vector3(0,1,0), Ogre::Radian(yaw_correction));
  focal_point = rot * focal_point;
  for(int i=0; i<3; i++) params[3+i] = focal_point[i];

  //set the camera params
  std::ostringstream os;
  for(int i=0; i<6; i++) os << params[i] << ' ';
  std::string s(os.str());
  vis_manager_->setTargetFrame("base_link");
  vis_manager_->setCurrentViewControllerType( "Orbit" );
  rviz::ViewController* vc = vis_manager_->getCurrentViewController();
  vc->fromString(s);
  vis_manager_->queueRender();
}

void InteractiveManipulationFrontend::cameraLeftButtonClicked( wxCommandEvent& event )
{
  setCamera( object_manipulator::cameraConfigurations().get_camera_pose("left") );
}

void InteractiveManipulationFrontend::cameraRightButtonClicked( wxCommandEvent& event )
{
  setCamera( object_manipulator::cameraConfigurations().get_camera_pose("right") );
}

void InteractiveManipulationFrontend::cameraFrontButtonClicked( wxCommandEvent& event )
{
  setCamera( object_manipulator::cameraConfigurations().get_camera_pose("front") );
}

void InteractiveManipulationFrontend::cameraTopButtonClicked( wxCommandEvent& event )
{
  setCamera( object_manipulator::cameraConfigurations().get_camera_pose("top") );
}

}
