#include <vfh_recognition/vfh_recognition.h>

#include "household_objects_database/database_view.h"
#include "household_objects_database/database_vfh.h"
#include "household_objects_database/objects_database.h"
#include "household_objects_database/database_helper_classes.h"

#include <vfh_recognizer_db/common_io.h>

//Class that implements the functionality of vfh_recognition::VFHRecognizer
//for the household_objects_database, creating kdtree from the db,
//loading views from the db, etc.

namespace vfh_recognizer_db
{
  template<template<class > class Distance>
  class VFHRecognizerDB : public vfh_recognition::VFHRecognizer<Distance>
  {
    typedef vfh_recognition::VFHRecognizer<Distance> base_t;
    using base_t::models_;
    using base_t::data_;
    using base_t::index_;

    typedef Distance<float> DistT;

    public:
      VFHRecognizerDB (bool enable_cache = false) : cache_enabled_(enable_cache) {}
      bool isCacheEnabled () {return cache_enabled_;}
      void setCacheEnabled (bool e) {cache_enabled_ = e;}

    private:

      household_objects_database::ObjectsDatabase * database;

      bool
      getPointCloudFromId (pcl::PointCloud<pcl::PointNormal> & cloud, std::string id);

      bool
      getVFHRollOrientationFromId (pcl::PointCloud<pcl::VFHSignature308> & vfh_orientation_signature, std::string id);

      bool
      getVFHRollOrientationFromIdThroughView (pcl::PointCloud<pcl::VFHSignature308> & vfh_orientation_signature, std::string id);

      bool
      getVFHCentroidFromVFHid (std::vector<float> & centroid, std::string id);

      bool
      getViewCentroidFromVFHid (std::vector<float> & centroid, std::string id);

      bool
      getPoseFromId(std::string id, geometry_msgs::Pose & pose);

      std::string
      getModelId (std::string id);

      int
      getVFHId (std::string id);

    public:

      inline int buildTreeFromViews (std::vector<boost::shared_ptr<household_objects_database::DatabaseView> > & views);

      inline bool buildTreeFromDB (int iteration);
      inline bool buildCachesFromDB (bf::path dpath, int iteration);

      bool
      getVFHHistogramFromVFHId (pcl::PointCloud<pcl::VFHSignature308> & vfh_descriptor, std::string vfh_id);

      int initializeFromViews(std::vector<boost::shared_ptr<household_objects_database::DatabaseView> > & views);

      ////////////////////////////////////////////////////////////////////////////////
      /** \brief Initialize the data structures for recognition
       * \param useDB - if true creates the structures from the database
       *              - if false, uses the cached kdtree if found, otherwise builds from DB
       */

      bool
      initialize (bool useDB, bf::path dpath, int linear = 0, int iteration = 0, bool no_save = false);

      ////////////////////////////////////////////////////////////////////////////////
      /** \brief Given a cluster returns the n-most similar models in models_, their poses
       * and the detection confidence.
       */

      bool
      detect (const sensor_msgs::PointCloud2ConstPtr& msg, int nModels, std::vector<int>& model_ids, std::vector<
          geometry_msgs::Pose> & poses, std::vector<float>& confidences, bool use_fitness_score = true, std::vector<
          boost::shared_ptr<household_objects_database::DatabaseView> > * views = NULL, std::vector<std::string> * vfh_ids_vec = NULL);


      bool cache_enabled_;
      std::map <int, pcl::PointCloud<pcl::VFHSignature308>, std::less<int>, Eigen::aligned_allocator<pcl::VFHSignature308> > roll_cache_;
      std::map <int, pcl::PointCloud<pcl::VFHSignature308>, std::less<int>, Eigen::aligned_allocator<pcl::VFHSignature308> > vfh_cache_;
      std::map <int, pcl::PointCloud<pcl::PointNormal>, std::less<int>, Eigen::aligned_allocator<pcl::PointNormal> > view_cache_;
      std::map <int, std::vector<float> > cluster_centroids_cache_;
      std::map <int, geometry_msgs::Pose > pose_cache_;
      std::map <int, std::string > scaled_model_id_cache_;
  };

}
