/*
 * normals_vfh_db_estimator_service.cpp
 *
 *  Created on: Jan 25, 2011
 *      Author: aaldoma
 */

#include "ros/ros.h"
#include <ros/message_traits.h>
#include <ros/serialization.h>
#include "sensor_msgs/PointCloud2.h"
#include "geometry_msgs/Pose.h"
#include "vfh_recognizer_fs/normals_vfh_db_estimator.h"
#include <pcl/common/common.h>
#include <fstream>
#include <iostream>
#include <pcl_ros/point_cloud.h>
#include <vfh_recognizer_fs/vfh_recognition.h>
#include <pcl/io/pcd_io.h>
#include <boost/filesystem.hpp>
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <Eigen/StdVector>

vfh_recognizer_fs::VFHRecognizerFS<flann::ChiSquareDistance> vfh_recognizer;

using namespace std;
namespace ser = ros::serialization;
bool USE_CLUSTER_CENTROIDS_ = true;
bool NORMALIZE_VFH_BINS_ = false;
bool FILL_SIZE_COMPONENT_ = true;
bool USE_OLD_VFH_ = false;

bool
create_directories_fs (std::string root)
{
  //create views directory if does not exist
  boost::filesystem::path root_path = root;
  if (!boost::filesystem::exists (root_path))
  {
    boost::filesystem::create_directory (root_path);
  }

  boost::filesystem::path views_dir = root_path / "views";
  if (!boost::filesystem::exists (views_dir))
  {
    boost::filesystem::create_directory (views_dir);
  }

  boost::filesystem::path views_vfh_map_dir = root_path / "views_vfh_map";
  if (!boost::filesystem::exists (views_vfh_map_dir))
  {
    boost::filesystem::create_directory (views_vfh_map_dir);
  }

  boost::filesystem::path vfh_views_map_dir = root_path / "vfh_views_map";
   if (!boost::filesystem::exists (vfh_views_map_dir))
   {
     boost::filesystem::create_directory (vfh_views_map_dir);
   }

  boost::filesystem::path vfhs_dir = root_path / "vfhs";
  if (!boost::filesystem::exists (vfhs_dir))
  {
    boost::filesystem::create_directory (vfhs_dir);
  }

  boost::filesystem::path vfhs_roll_orientantion_dir = root_path / "vfhs_roll_orientation";
  if (!boost::filesystem::exists (vfhs_roll_orientantion_dir))
  {
    boost::filesystem::create_directory (vfhs_roll_orientantion_dir);
  }

  boost::filesystem::path transforms_dir = root_path / "transforms";
  if (!boost::filesystem::exists (transforms_dir))
  {
    boost::filesystem::create_directory (transforms_dir);
  }

  boost::filesystem::path centroids_dir = root_path / "centroids";
  if (!boost::filesystem::exists (centroids_dir))
  {
    boost::filesystem::create_directory (centroids_dir);
  }

  return true;
}

bool
writePoseToFile (std::string file, geometry_msgs::Pose pose)
{
  ofstream out (file.c_str ());
  if (!out)
  {
    cout << "Cannot open file.\n";
    return false;
  }

  out << pose.position.x << " " << pose.position.y << " " << pose.position.z << " " << pose.orientation.x << " "
      << pose.orientation.y << " " << pose.orientation.z << " " << pose.orientation.w << endl;
  out.close ();

  return true;
}

bool
writeViewVFHMapToFile (std::string file, std::vector<int> map)
{
  ofstream out (file.c_str ());
  if (!out)
  {
    cout << "Cannot open file.\n";
    return false;
  }

  std::stringstream value;
  for(size_t i=0; i < map.size(); i++) {
      value << map[i] << " ";
  }

  out << value.str().substr(0,-1);
  out.close ();

  return true;
}

bool
writeVFHViewMapToFile (std::string file, int view_id)
{
  ofstream out (file.c_str ());
  if (!out)
  {
    cout << "Cannot open file.\n";
    return false;
  }

  out << view_id;
  out.close ();

  return true;
}

bool
writeCentroidToFile (std::string file, Eigen::Vector3f centroid)
{
  ofstream out (file.c_str ());
  if (!out)
  {
    cout << "Cannot open file.\n";
    return false;
  }

  out << centroid[0] << " " << centroid[1] << " " << centroid[2] << endl;
  out.close ();

  return true;
}

int VFH_id_ = 0;
int VIEW_id_ = 0;

bool
compute_and_store (vfh_recognizer_fs::normals_vfh_db_estimator::Request &req,
                   vfh_recognizer_fs::normals_vfh_db_estimator::Response &res)
{
  pcl::PointCloud<pcl::PointNormal>::Ptr cloud_normals (new pcl::PointCloud<pcl::PointNormal> ());
  pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_signaturePtr (new pcl::PointCloud<pcl::VFHSignature308> ());
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ> ());
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_noise (new pcl::PointCloud<pcl::PointXYZ> ());
  pcl::PointCloud<pcl::PointNormal>::Ptr cloud_normals_no_noise (new pcl::PointCloud<pcl::PointNormal> ());
  pcl::PointCloud < pcl::VFHSignature308 > vfh_signature;

  pcl::fromROSMsg (req.view, *cloud);

  //check if cloud has any points
  if (cloud->points.size () == 0)
  {
    ROS_WARN ("Cloud has no points, skip view!");
    return false;
  }

  pcl::copyPointCloud (*cloud, *cloud_noise);

  //Add noise to training data so we can use same parameters as for real data
  bool noisify = true;

  if (noisify)
  {
    double noise_std = 0.001; //1mm

    struct timeval start;
    gettimeofday (&start, NULL);

    boost::mt19937 rng;
    rng.seed (start.tv_usec);
    boost::normal_distribution<> nd (0.0, noise_std);
    boost::variate_generator<boost::mt19937&, boost::normal_distribution<> > var_nor (rng, nd);

    // Noisify each point in the dataset
    for (size_t cp = 0; cp < cloud_noise->points.size (); ++cp)
    {
      cloud_noise->points[cp].z += var_nor ();
    }
  }

  std::vector<pcl::PointCloud<pcl::VFHSignature308>,Eigen::aligned_allocator< pcl::PointCloud<pcl::VFHSignature308> > > vfh_signatures;
  std::vector < Eigen::Vector3f > centroids_dominant_orientations;

  //CONFIGURATION TO USE CLUSTERS CENTROIDS OR VIEW CENTROIDS
  vfh_recognizer.setUseClusterCentroids (USE_CLUSTER_CENTROIDS_);

  if (!NORMALIZE_VFH_BINS_)
  {
    std::cout << "BINS WONT BE NORMALIZED" << std::endl;
    vfh_recognizer.setNormalizeVFHBins (false);
  }

  vfh_recognizer.setUseOldVFH (USE_OLD_VFH_);
  vfh_recognizer.fill_size_component_ = FILL_SIZE_COMPONENT_;

  struct timeval start;
  gettimeofday (&start, NULL);

  vfh_recognizer.view_id_being_processed_ = start.tv_usec;

  vfh_recognizer.computeVFH (cloud_noise, cloud_normals, vfh_signatures, centroids_dominant_orientations);
  vfh_recognizer.computeNormals (cloud, cloud_normals_no_noise);

  vfh_signaturePtr = vfh_signatures[0].makeShared ();

  //Check the size of the normals and of the vfh_signature
  if ((cloud_normals->points.size () < 100) || (ser::serializationLength (*vfh_signaturePtr) < (sizeof(float) * 308)))
  {
    ROS_WARN ("Normals size or vfh descriptor size are wrong, skip this view - model_id:");
    return false;
  }

  create_directories_fs (req.root_dir.data);

  std::vector<int> ids_for_this_view;

  for (size_t vfh_idx = 0; vfh_idx < vfh_signatures.size (); vfh_idx++)
  {
    std::vector<float> vfh_centroid (3, 0);
    for (size_t i = 0; i < 3; i++)
      vfh_centroid[i] = centroids_dominant_orientations[vfh_idx][i];

    if (USE_CLUSTER_CENTROIDS_)
    {
      //Once the VFH has been estimated, we need to compute the orientation histogram
      //for each VFH using the cluster centroid
      pcl::PointCloud<pcl::VFHSignature308>::Ptr
                                                 vfh_orientation_signature (
                                                                            new pcl::PointCloud<pcl::VFHSignature308> ());
      cv::Mat frequency_domain;
      vfh_recognizer.computeVFHRollOrientation (cloud_normals, vfh_orientation_signature,
                                                centroids_dominant_orientations[vfh_idx], frequency_domain); //use the centroid of the dominant clusters...

      //SAVE vfh_signature, vfh_orientation_signature, cloud_normals and req.transform
      stringstream path_vfh;
      path_vfh << req.root_dir.data << "/vfhs/vfh_" << req.model_id.data << "_" << VFH_id_ << ".pcd";
      std::cout << path_vfh.str () << std::endl;
      pcl::io::savePCDFileBinary (path_vfh.str (), vfh_signatures[vfh_idx]);

      stringstream path_vfh_orientation;
      path_vfh_orientation << req.root_dir.data << "/vfhs_roll_orientation/vfh_orientation_" << req.model_id.data << "_"
          << VFH_id_ << ".pcd";
      std::cout << path_vfh_orientation.str () << std::endl;
      pcl::io::savePCDFileBinary (path_vfh_orientation.str (), *vfh_orientation_signature);

      stringstream path_transform;
      path_transform << req.root_dir.data << "/transforms/transform_" << req.model_id.data << "_" << VFH_id_
          << ".txt";
      std::cout << path_transform.str () << std::endl;
      writePoseToFile (path_transform.str (), req.transform);

      stringstream path_centroid;
      path_centroid << req.root_dir.data << "/centroids/centroid_" << req.model_id.data << "_" << VFH_id_
          << ".txt";
      std::cout << path_centroid.str () << std::endl;
      writeCentroidToFile (path_centroid.str (), centroids_dominant_orientations[vfh_idx]);

      stringstream path_vfh_view_map;
      path_vfh_view_map << req.root_dir.data << "/vfh_views_map/vfh_view_" << req.model_id.data << "_" << VFH_id_
              << ".txt";

      writeVFHViewMapToFile(path_vfh_view_map.str(),VIEW_id_);

      ids_for_this_view.push_back(VFH_id_);
      VFH_id_++;
     }
  }

  stringstream path_view_vfh_map;
  path_view_vfh_map << req.root_dir.data << "/views_vfh_map/view_vfh_" << req.model_id.data << "_" << VIEW_id_
          << ".txt";

  writeViewVFHMapToFile(path_view_vfh_map.str(),ids_for_this_view);

  stringstream path_view;
  path_view << req.root_dir.data << "/views/view_" << req.model_id.data << "_" << VIEW_id_ << ".pcd";
  std::cout << path_view.str () << std::endl;
  pcl::io::savePCDFileBinary (path_view.str (), *cloud_normals);
  VIEW_id_++;

  /*if (!USE_CLUSTER_CENTROIDS_)
  {
    //Once the VFH has been estimated, we need to compute the orientation histogram
    pcl::PointCloud < pcl::PointNormal > grid;
    grid = *cloud_normals_no_noise; //TODO
    Eigen::Vector4f centroidView;
    pcl::compute3DCentroid (grid, centroidView);
    pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_orientation_signature (new pcl::PointCloud<pcl::VFHSignature308> ());
    Eigen::Vector3f ci3f (centroidView[0], centroidView[1], centroidView[2]);
    cv::Mat frequency_domain;
    vfh_recognizer.computeVFHRollOrientation (cloud_normals, vfh_orientation_signature, ci3f, frequency_domain); //use the centroid of the view

    //SAVE vfh_signature, vfh_orientation_signature, cloud_normals and req.transform
    stringstream path_view;
    path_view << req.root_dir.data << "/views/view_" << req.model_id.data << "_" << req.view_id.data << ".pcd";
    std::cout << path_view.str () << std::endl;
    pcl::io::savePCDFileBinary (path_view.str (), *cloud_normals);

    stringstream path_vfh;
    path_vfh << req.root_dir.data << "/vfhs/vfh_" << req.model_id.data << "_" << req.view_id.data << ".pcd";
    std::cout << path_vfh.str () << std::endl;
    pcl::io::savePCDFileBinary (path_vfh.str (), vfh_signature);

    stringstream path_vfh_orientation;
    path_vfh_orientation << req.root_dir.data << "/vfhs_roll_orientation/vfh_orientation_" << req.model_id.data << "_"
        << req.view_id.data << ".pcd";
    std::cout << path_vfh_orientation.str () << std::endl;
    pcl::io::savePCDFileBinary (path_vfh_orientation.str (), *vfh_orientation_signature);

    stringstream path_transform;
    path_transform << req.root_dir.data << "/transforms/transform_" << req.model_id.data << "_" << req.view_id.data
        << ".txt";
    std::cout << path_transform.str () << std::endl;
    writePoseToFile (path_transform.str (), req.transform);

    stringstream path_centroid;
    path_centroid << req.root_dir.data << "/transforms/transform_" << req.model_id.data << "_" << req.view_id.data
        << ".txt";
    std::cout << path_centroid.str () << std::endl;
    writeCentroidToFile (path_centroid.str (), centroids_dominant_orientations[0]);
  }*/

  //pcl::io::savePCDFileBinary (path_vfh_orientation.str(), *vfh_orientation_signature);

  return true;
}

int
main (int argc, char **argv)
{
  ros::init (argc, argv, "normals_vfh_db_estimator_server");
  ros::NodeHandle n;

  ros::ServiceServer service = n.advertiseService ("normals_vfh_db_estimator", compute_and_store);
  ros::spin ();

  return 0;
}
